@@ -23,8 +23,8 @@ import logging
 import tempfile
 import threading
 import subprocess
-from errno import EEXIST, ENOENT, ENODATA, ENOTDIR, ELOOP
-from errno import EISDIR, ENOTEMPTY, ESTALE, EINVAL
+from errno import EEXIST, ENOENT, ENODATA, ENOTDIR, ELOOP, EACCES
+from errno import EISDIR, ENOTEMPTY, ESTALE, EINVAL, EBUSY, EPERM
 from select import error as SelectError
 import shutil
 
@@ -41,6 +41,7 @@ from syncdutils import get_changelog_log_level
 from syncdutils import CHANGELOG_AGENT_CLIENT_VERSION
 from gsyncdstatus import GeorepStatus
 from syncdutils import get_master_and_slave_data_from_args
+from syncdutils import mntpt_list
 
 UrlRX = re.compile('\A(\w+)://([^ *?[]*)\Z')
 HostRX = re.compile('[a-zA-Z\d](?:[a-zA-Z\d.-]*[a-zA-Z\d])?', re.I)
@@ -643,7 +644,7 @@ class Server(object):
                                st['uid'], st['gid'],
                                gf, st['mode'], bn, lnk)
 
-        def entry_purge(entry, gfid):
+        def entry_purge(op, entry, gfid):
             # This is an extremely racy code and needs to be fixed ASAP.
             # The GFID check here is to be sure that the pargfid/bname
             # to be purged is the GFID gotten from the changelog.
@@ -652,28 +653,32 @@ class Server(object):
             if not matching_disk_gfid(gfid, entry):
                 return
 
-            er = errno_wrap(os.unlink, [entry], [ENOENT, ESTALE, EISDIR])
-            if isinstance(er, int):
-                if er == EISDIR:
-                    er = errno_wrap(os.rmdir, [entry], [ENOENT, ESTALE,
-                                                        ENOTEMPTY])
-                    if er == ENOTEMPTY:
-                        return er
+            if op == 'UNLINK':
+                errno_wrap(os.unlink, [entry], [ENOENT, ESTALE],
+                           [EBUSY])
+            elif op == 'RMDIR':
+                er = errno_wrap(os.rmdir, [entry], [ENOENT, ESTALE,
+                                                    ENOTEMPTY], [EBUSY])
+                if er == ENOTEMPTY:
+                    return er
 
         def collect_failure(e, cmd_ret):
             # We do this for failing fops on Slave
             # Master should be logging this
             if cmd_ret is None:
-                return
+                return False
 
             if cmd_ret == EEXIST:
                 disk_gfid = cls.gfid_mnt(e['entry'])
-                if isinstance(disk_gfid, basestring):
-                    if e['gfid'] != disk_gfid:
-                        failures.append((e, cmd_ret, disk_gfid))
+                if isinstance(disk_gfid, basestring) and e['gfid'] != disk_gfid:
+                    failures.append((e, cmd_ret, disk_gfid))
+                else:
+                    return False
             else:
                 failures.append((e, cmd_ret))
 
+            return True
+
         failures = []
 
         def matching_disk_gfid(gfid, entry):
@@ -708,7 +713,7 @@ class Server(object):
                 if not matching_disk_gfid(gfid, entry):
                     return
                 er = errno_wrap(os.remove, [fullname], [ENOENT, ESTALE,
-                                                        EISDIR])
+                                                        EISDIR], [EBUSY])
 
                 if er == EISDIR:
                     recursive_rmdir(gfid, entry, fullname)
@@ -716,7 +721,7 @@ class Server(object):
             if not matching_disk_gfid(gfid, entry):
                 return
 
-            errno_wrap(os.rmdir, [path], [ENOENT, ESTALE])
+            errno_wrap(os.rmdir, [path], [ENOENT, ESTALE], [EBUSY])
 
         def rename_with_disk_gfid_confirmation(gfid, entry, en):
             if not matching_disk_gfid(gfid, entry):
@@ -729,7 +734,7 @@ class Server(object):
 
             cmd_ret = errno_wrap(os.rename,
                                  [entry, en],
-                                 [ENOENT, EEXIST], [ESTALE])
+                                 [ENOENT, EEXIST], [ESTALE, EBUSY])
             collect_failure(e, cmd_ret)
 
 
@@ -752,7 +757,7 @@ class Server(object):
             if op in ['RMDIR', 'UNLINK']:
                 # Try once, if rmdir failed with ENOTEMPTY
                 # then delete recursively.
-                er = entry_purge(entry, gfid)
+                er = entry_purge(op, entry, gfid)
                 if isinstance(er, int):
                     if er == ENOTEMPTY and op == 'RMDIR':
                         # Retry if ENOTEMPTY, ESTALE
@@ -819,12 +824,12 @@ class Server(object):
                             # we have a hard link, we can now unlink source
                             try:
                                 errno_wrap(os.unlink, [entry],
-                                           [ENOENT, ESTALE])
+                                           [ENOENT, ESTALE], [EBUSY])
                             except OSError as e:
                                 if e.errno == EISDIR:
                                     try:
                                         errno_wrap(os.rmdir, [entry],
-                                                   [ENOENT, ESTALE])
+                                                   [ENOENT, ESTALE], [EBUSY])
                                     except OSError as e:
                                         if e.errno == ENOTEMPTY:
                                             logging.error(
@@ -842,8 +847,16 @@ class Server(object):
                 cmd_ret = errno_wrap(Xattr.lsetxattr,
                                      [pg, 'glusterfs.gfid.newfile', blob],
                                      [EEXIST, ENOENT],
-                                     [ESTALE, EINVAL])
-                collect_failure(e, cmd_ret)
+                                     [ESTALE, EINVAL, EBUSY])
+                failed = collect_failure(e, cmd_ret)
+
+                # If directory creation is failed, return immediately before
+                # further processing. Allowing it to further process will
+                # cause the entire directory tree to fail syncing to slave.
+                # Hence master will log and raise exception if it's
+                # directory failure.
+                if failed and op == 'MKDIR':
+                    return failures
 
                 # If UID/GID is different than zero that means we are trying
                 # create Entry with different UID/GID. Create Entry with
@@ -852,7 +865,7 @@ class Server(object):
                     path = os.path.join(pfx, gfid)
                     cmd_ret = errno_wrap(os.chown, [path, uid, gid], [ENOENT],
                                          [ESTALE, EINVAL])
-                collect_failure(e, cmd_ret)
+                    collect_failure(e, cmd_ret)
 
         return failures
 
@@ -867,17 +880,43 @@ class Server(object):
             atime = e['stat']['atime']
             mtime = e['stat']['mtime']
             go = e['go']
-            cmd_ret = errno_wrap(os.chmod, [go, mode],
-                                 [ENOENT], [ESTALE, EINVAL])
-            # This is a fail fast mechanism
-            # We do this for failing fops on Slave
-            # Master should be logging this
+            # Linux doesn't support chmod on symlink itself.
+            # It is always applied to the target file. So
+            # changelog would record target file's gfid
+            # and we are good. But 'chown' is supported on
+            # symlink file. So changelog would record symlink
+            # gfid in such cases. Since we do 'chown' 'chmod'
+            # 'utime' for each gfid recorded for metadata, and
+            # we know from changelog the metadata is on symlink's
+            # gfid or target file's gfid, we should be doing
+            # 'lchown' 'lchmod' 'utime with no-deference' blindly.
+            # But since 'lchmod' and 'utime with no de-reference' is
+            # not supported in python3, we have to rely on 'chmod'
+            # and 'utime with de-reference'. But 'chmod'
+            # de-reference the symlink and gets ENOENT, EACCES,
+            # EPERM errors, hence ignoring those errors if it's on
+            # symlink file.
+
+            is_symlink = False
+            cmd_ret = errno_wrap(os.lchown, [go, uid, gid], [ENOENT],
+                                 [ESTALE, EINVAL])
             if isinstance(cmd_ret, int):
-                failures.append((e, cmd_ret))
                 continue
-            errno_wrap(os.chown, [go, uid, gid], [ENOENT], [ESTALE, EINVAL])
-            errno_wrap(os.utime, [go, (atime, mtime)],
-                       [ENOENT], [ESTALE, EINVAL])
+
+            cmd_ret = errno_wrap(os.chmod, [go, mode],
+                                 [ENOENT, EACCES, EPERM], [ESTALE, EINVAL])
+            if isinstance(cmd_ret, int):
+                is_symlink = os.path.islink(go)
+                if not is_symlink:
+                    failures.append((e, cmd_ret, "chmod"))
+
+            cmd_ret = errno_wrap(os.utime, [go, (atime, mtime)],
+                                 [ENOENT, EACCES, EPERM], [ESTALE, EINVAL])
+            if isinstance(cmd_ret, int):
+                if not is_symlink:
+                    is_symlink = os.path.islink(go)
+                if not is_symlink:
+                    failures.append((e, cmd_ret, "utime"))
         return failures
 
     @classmethod
@@ -995,8 +1034,16 @@ class SlaveRemote(object):
         so = getattr(gconf, 'session_owner', None)
         if so:
             extra_opts += ['--session-owner', so]
+        li = getattr(gconf, 'local_id', None)
+        if li:
+            extra_opts += ['--local-id', li]
+        ln = getattr(gconf, 'local_node', None)
+        if ln:
+            extra_opts += ['--local-node', ln]
         if boolify(gconf.use_rsync_xattrs):
             extra_opts.append('--use-rsync-xattrs')
+        if boolify(gconf.access_mount):
+            extra_opts.append('--access-mount')
         po = Popen(rargs + gconf.remote_gsyncd.split() + extra_opts +
                    ['-N', '--listen', '--timeout', str(gconf.timeout),
                     slave],
@@ -1037,8 +1084,10 @@ class SlaveRemote(object):
             (boolify(gconf.sync_acls) and ['--acls'] or []) + \
             ['.'] + list(args)
 
-        if boolify(gconf.configinterface.get_realtime(
-                "log_rsync_performance")):
+        log_rsync_performance = boolify(gconf.configinterface.get_realtime(
+            "log_rsync_performance", default_value=False))
+
+        if log_rsync_performance:
             # use stdout=PIPE only when log_rsync_performance enabled
             # Else rsync will write to stdout and nobody is their
             # to consume. If PIPE is full rsync hangs.
@@ -1057,8 +1106,7 @@ class SlaveRemote(object):
             for errline in stderr.strip().split("\n")[:-1]:
                 logging.error("SYNC Error(Rsync): %s" % errline)
 
-        if boolify(gconf.configinterface.get_realtime(
-                "log_rsync_performance")):
+        if log_rsync_performance:
             rsync_msg = []
             for line in stdout.split("\n"):
                 if line.startswith("Number of files:") or \
@@ -1355,13 +1403,14 @@ class GLUSTER(AbstractUrl, SlaveLocal, SlaveRemote):
                         assert(mntdata[-1] == '\0')
                         mntpt = mntdata[:-1]
                         assert(mntpt)
-                        if mounted:
+                        if mounted and not boolify(gconf.access_mount):
                             po = self.umount_l(mntpt)
                             po.terminate_geterr(fail_on_err=False)
                             if po.returncode != 0:
                                 po.errlog()
                                 rv = po.returncode
-                        self.cleanup_mntpt(mntpt)
+                        if not boolify(gconf.access_mount):
+                            self.cleanup_mntpt(mntpt)
                 except:
                     logging.exception('mount cleanup failure:')
                     rv = 200
@@ -1381,13 +1430,14 @@ class GLUSTER(AbstractUrl, SlaveLocal, SlaveRemote):
 
         def make_mount_argv(self):
             self.mntpt = tempfile.mkdtemp(prefix='gsyncd-aux-mount-')
+            mntpt_list.append(self.mntpt)
             return [self.get_glusterprog()] + \
                 ['--' + p for p in self.params] + [self.mntpt]
 
         def cleanup_mntpt(self, mntpt=None):
             if not mntpt:
                 mntpt = self.mntpt
-            os.rmdir(mntpt)
+            errno_wrap(os.rmdir, [mntpt], [ENOENT, EBUSY])
 
     class MountbrokerMounter(Mounter):
 
@@ -1428,6 +1478,8 @@ class GLUSTER(AbstractUrl, SlaveLocal, SlaveRemote):
         with given backend
         """
 
+        logging.info ("Mounting gluster volume locally...")
+        t0 = time.time()
         label = getattr(gconf, 'mountbroker', None)
         if not label and not privileged():
             label = syncdutils.getusername()
@@ -1438,6 +1490,8 @@ class GLUSTER(AbstractUrl, SlaveLocal, SlaveRemote):
             ['log-file=' + gconf.gluster_log_file, 'volfile-server=' +
              self.host, 'volfile-id=' + self.volume, 'client-pid=-1']
         mounter(params).inhibit(*[l for l in [label] if l])
+        logging.info ("Mounted gluster volume. Time taken: {0:.4f} "
+                      "secs".format((time.time() - t0)))
 
     def connect_remote(self, *a, **kw):
         sup(self, *a, **kw)
@@ -1547,7 +1601,9 @@ class GLUSTER(AbstractUrl, SlaveLocal, SlaveRemote):
             changelog_agent = RepceClient(int(inf), int(ouf))
             master_name, slave_data = get_master_and_slave_data_from_args(
                 sys.argv)
-            status = GeorepStatus(gconf.state_file, gconf.local_path,
+            status = GeorepStatus(gconf.state_file, gconf.local_node,
+                                  gconf.local_path,
+                                  gconf.local_node_id,
                                   master_name, slave_data)
             status.reset_on_worker_start()
             rv = changelog_agent.version()
@@ -1697,10 +1753,14 @@ class SSH(AbstractUrl, SlaveRemote):
                                  self.inner_rsc.url)
 
         deferred = go_daemon == 'postconn'
+        logging.info ("Initializing SSH connection between master and slave...")
+        t0 = time.time()
         ret = sup(self, gconf.ssh_command.split() +
                   ["-p", str(gconf.ssh_port)] +
                   gconf.ssh_ctl_args + [self.remote_addr],
                   slave=self.inner_rsc.url, deferred=deferred)
+        logging.info ("SSH connection between master and slave established. "
+                      "Time taken: {0:.4f} secs".format((time.time() - t0)))
 
         if deferred:
             # send a message to peer so that we can wait for
